@@ -31,24 +31,35 @@ def test_command(tmpdir):
 def test_basic(tmpdir, capsys):
     """Test some basic functionality"""
     assert cs.main('_does_not_exist_') == 0
-    with open(op.join(str(tmpdir), 'tmp'), 'w') as f:
+    fname = op.join(str(tmpdir), 'tmp')
+    with open(fname, 'w') as f:
         pass
     assert cs.main('-D', 'foo', f.name) == 1, 'missing dictionary'
-    try:
-        assert 'cannot find dictionary' in capsys.readouterr()[1]
-        assert cs.main(f.name) == 0, 'empty file'
-        with open(f.name, 'a') as f:
-            f.write('this is a test file\n')
-        assert cs.main(f.name) == 0, 'good'
-        with open(f.name, 'a') as f:
-            f.write('abandonned\n')
-        assert cs.main(f.name) == 1, 'bad'
-        with open(f.name, 'a') as f:
-            f.write('abandonned\n')
-        assert cs.main(f.name) == 2, 'worse'
-    finally:
-        os.remove(f.name)
+    assert 'cannot find dictionary' in capsys.readouterr()[1]
+    assert cs.main(fname) == 0, 'empty file'
+    with open(fname, 'a') as f:
+        f.write('this is a test file\n')
+    assert cs.main(fname) == 0, 'good'
+    with open(fname, 'a') as f:
+        f.write('abandonned\n')
+    assert cs.main(fname) == 1, 'bad'
+    with open(fname, 'a') as f:
+        f.write('abandonned\n')
+    assert cs.main(fname) == 2, 'worse'
+    with open(fname, 'a') as f:
+        f.write('tim\ngonna\n')
+    assert cs.main(fname) == 2, 'with a name'
+    assert cs.main('--builtin', 'clear,rare,names,informal', fname) == 4
+    capsys.readouterr()
+    assert cs.main(fname, '--builtin', 'foo') == 1  # bad type sys.exit(1)
+    stdout = capsys.readouterr()[1]
+    assert 'Unknown builtin dictionary' in stdout
     d = str(tmpdir)
+    assert cs.main(fname, '-D', op.join(d, 'foo')) == 1  # bad dict
+    stdout = capsys.readouterr()[1]
+    assert 'cannot find dictionary' in stdout
+    os.remove(fname)
+
     with open(op.join(d, 'bad.txt'), 'w') as f:
         f.write('abandonned\nAbandonned\nABANDONNED\nAbAnDoNnEd')
     assert cs.main(d) == 4
@@ -302,6 +313,65 @@ def test_case_handling(tmpdir, capsys):
         os.remove(f.name)
 
 
+def test_context(tmpdir, capsys):
+    """Test context options."""
+    d = str(tmpdir)
+    with open(op.join(d, 'context.txt'), 'w') as f:
+        f.write('line 1\nline 2\nline 3 abandonned\nline 4\nline 5')
+
+    # symmetric context, fully within file
+    cs.main('-C', '1', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert len(lines) == 5
+    assert lines[0] == ': line 2'
+    assert lines[1] == '> line 3 abandonned'
+    assert lines[2] == ': line 4'
+
+    # requested context is bigger than the file
+    cs.main('-C', '10', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert len(lines) == 7
+    assert lines[0] == ': line 1'
+    assert lines[1] == ': line 2'
+    assert lines[2] == '> line 3 abandonned'
+    assert lines[3] == ': line 4'
+    assert lines[4] == ': line 5'
+
+    # only before context
+    cs.main('-B', '2', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert len(lines) == 5
+    assert lines[0] == ': line 1'
+    assert lines[1] == ': line 2'
+    assert lines[2] == '> line 3 abandonned'
+
+    # only after context
+    cs.main('-A', '1', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert len(lines) == 4
+    assert lines[0] == '> line 3 abandonned'
+    assert lines[1] == ': line 4'
+
+    # asymmetric context
+    cs.main('-B', '2', '-A', '1', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert len(lines) == 6
+    assert lines[0] == ': line 1'
+    assert lines[1] == ': line 2'
+    assert lines[2] == '> line 3 abandonned'
+    assert lines[3] == ': line 4'
+
+    # both '-C' and '-A' on the command line
+    cs.main('-C', '2', '-A', '1', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert 'ERROR' in lines[0]
+
+    # both '-C' and '-B' on the command line
+    cs.main('-C', '2', '-B', '1', d)
+    lines = capsys.readouterr()[0].split('\n')
+    assert 'ERROR' in lines[0]
+
+
 @contextlib.contextmanager
 def FakeStdin(text):
     if sys.version[0] == '2':
